/*  $Header: /home/cvsroot/dvipdfmx/src/fontmap.c,v 1.20 2004/08/22 10:26:24 hirata Exp $
    
    This is dvipdfmx, an eXtended version of dvipdfm by Mark A. Wicks.

    Copyright (C) 2002 by Jin-Hwan Cho and Shunsaku Hirata,
    the dvipdfmx project team <dvipdfmx@project.ktug.or.kr>
    
    Copyright (C) 1998, 1999 by Mark A. Wicks <mwicks@kettering.edu>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.
    
    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.
    
    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA.
*/

#if HAVE_CONFIG_H
#include "config.h"
#endif

#include "mem.h"
#include "error.h"
#include "dpxutil.h"
#include "pdfparse.h"
#include "pdflimits.h"
#include "subfont.h"

#include "dpxfile.h"
#include "otl_opt.h"
#include "fontmap.h"

/* CIDFont */
static char *strip_options (const char *map_name, fontmap_opt *opt);

static int verbose = 0;

void
fontmap_set_verbose (void)
{
  verbose++;
}

static int num_fontmap = 0, max_fontmap = 0;

static struct map_record {
  char *tex_name;

  char *font_name;
  char *enc_name;
  char *sfd_name;

  fontmap_opt opt;

  int deleted;
} *fontmap = NULL;

#define FONTMAP_ALLOC_SIZE 64

static void
init_map_record (struct map_record *mrec) 
{
  mrec->tex_name  = NULL;
  mrec->enc_name  = NULL;
  mrec->font_name = NULL;
  mrec->sfd_name  = NULL;

  mrec->opt.slant  = 0.0;
  mrec->opt.extend = 1.0;
  mrec->opt.bold   = 0.0;
  mrec->opt.flags  = 0;
  mrec->opt.mapc   = 0;

  mrec->opt.otl_tags = NULL;
  mrec->opt.index    = 0;
  mrec->opt.charcoll = NULL;
  mrec->opt.style    = FONTMAP_STYLE_NONE;

  mrec->deleted = 0;
}

static void
clear_map_record (struct map_record *mrec) 
{
  if (!mrec)
    return;

  if (mrec->tex_name)  RELEASE(mrec->tex_name);
  if (mrec->enc_name)  RELEASE(mrec->enc_name);
  if (mrec->sfd_name)  RELEASE(mrec->sfd_name);
  if (mrec->font_name) RELEASE(mrec->font_name);

  if (mrec->opt.otl_tags) RELEASE(mrec->opt.otl_tags);
  if (mrec->opt.charcoll) RELEASE(mrec->opt.charcoll);

  init_map_record(mrec);
}

#define GET_MAP_RECORD(n) (((n) < num_fontmap) ? &(fontmap[(n)]) : NULL)
#define ENABLE_MAP_RECORD(r) {\
  (r)->deleted = 0;\
}
#define DISABLE_MAP_RECORD(r) {\
  (r)->deleted = 1;\
}

void
release_map_record (void)
{
  struct map_record *mrec;
  int map_id;

  if (fontmap) {
    for (map_id = 0; map_id < num_fontmap; map_id++) {
      mrec = GET_MAP_RECORD(map_id);
      clear_map_record(mrec);
    }
    RELEASE(fontmap);
  }
  fontmap     = NULL;
  num_fontmap = 0;
}

static void
fill_in_defaults (struct map_record *mrec)
{
  char *p;
  int   len;

  if (!mrec->tex_name)
    ERROR("Font map entry without tex name? You found a bug in fontmap.c...");

  if (mrec->enc_name && 
      (!strcmp(mrec->enc_name, "default") || !strcmp(mrec->enc_name, "none"))) {
    RELEASE(mrec->enc_name);
    mrec->enc_name = NULL;
  }
  if (mrec->font_name && 
      (!strcmp(mrec->font_name, "default") || !strcmp(mrec->font_name, "none"))) {
    RELEASE(mrec->font_name);
    mrec->font_name = NULL;
  }

  /* We *must* fill font_name either explicitly or by default */
  if (mrec->font_name == NULL) {
    /* In the case of subfonts, the base name (before the character '@')
     * will be used as a font_name by default.
     * Otherwise tex_name will be used as a font_name by default. */
    if (mrec->sfd_name) {
      p = strchr(mrec->tex_name, '@');
      if (!p || p[1] == '\0' || p == mrec->tex_name)
	ERROR("Expecting subfont but no sfd/fontname found!!!");
      len = strlen(mrec->tex_name) - strlen(p);
      mrec->font_name = NEW(len+1, char);
      strncpy(mrec->font_name, mrec->tex_name, len);
      (mrec->font_name)[len] = '\0';
    } else {
      mrec->font_name = NEW(strlen(mrec->tex_name)+1, char);
      strcpy(mrec->font_name, mrec->tex_name);
    }
  }
}

static struct map_record *
find_map_record (const char *tex_name)
{
  struct map_record *mrec;
  int    map_id;
  int    name_len, base_len, sfdn_len;

  for (map_id = 0; map_id < num_fontmap; map_id++) {
    mrec = GET_MAP_RECORD(map_id);
    if (!mrec || !mrec->tex_name)
      ERROR("No TFM name?: id=%d (%d)", map_id, num_fontmap);
    else if (!mrec->sfd_name) {
      if (!strcmp(tex_name, mrec->tex_name))
	return mrec;
    } else {
      name_len = strlen(tex_name);
      base_len = strlen(mrec->tex_name);
      sfdn_len = strlen(mrec->sfd_name);
      if (name_len > base_len + sfdn_len &&
	  tex_name[base_len] == '@' &&
	  !strncmp(tex_name, mrec->tex_name, base_len)) {
	if (!strncmp(tex_name+base_len+1, mrec->sfd_name, sfdn_len) &&
	    ((name_len == base_len + sfdn_len + 1) ||
	     (name_len == base_len + sfdn_len + 2 &&
	      tex_name[name_len-1] == '@'))
	    )
	  return mrec;
      }
    }
  }

  return NULL;
}


int
fontmap_add_maprec (const char *tex_name,
		    char *mdef, char *endptr, int mode)
{
  struct map_record *mrec;
  char *p;
  int   is_new_entry = 1;

  /*
   * Parse record line in map file.  First two fields (after TeX font
   * name) are position specific.  Arguments start at the first token
   * beginning with a  '-'.
   */
  if (!tex_name)
    return -1;

  mrec = find_map_record(tex_name);
  switch (mode) {
  case FONTMAP_RMODE_APPEND: /* Do not replace. */
    if (mrec)
      return 0;
    break;
  case FONTMAP_RMODE_REMOVE:
    if (mrec)
      DISABLE_MAP_RECORD(mrec);
    return 0;
    break;
  }

  /* Replace or add */
  if (mrec) {
    clear_map_record(mrec);
    is_new_entry = 0;
  } else {
    if (num_fontmap >= max_fontmap) {
      max_fontmap += FONTMAP_ALLOC_SIZE;
      fontmap = RENEW(fontmap, max_fontmap, struct map_record);
    }
    mrec = &(fontmap[num_fontmap]);
    init_map_record(mrec);
    is_new_entry = 1;
  }

  /*
   * Subfonts are identified with the tex_name having
   * the character '@'.
   */
  p = strchr(tex_name, '@');
  if (p && p[1] != '\0') {
    char *sfd_name;

    p[0] = '\0';
    sfd_name = p + 1;
    if ((p = strstr(sfd_name, ".sfd")) != NULL)
      *p = '\0';
    else if ((p = strchr(sfd_name, '@')) != NULL)
      *p = '\0';
    mrec->sfd_name = NEW(strlen(sfd_name)+1, char);
    strcpy(mrec->sfd_name, sfd_name);
  }
  mrec->tex_name = NEW(strlen(tex_name)+1, char);
  strcpy(mrec->tex_name, tex_name);

  skip_white(&mdef, endptr);
  if (*mdef != '-') { /* May be NULL */
    mrec->enc_name = parse_ident(&mdef, endptr);
    skip_white(&mdef, endptr);
  }
  if (*mdef != '-') { /* May be NULL */
    /* parse_val_ident() allows inclusion of '/' */
    mrec->font_name = parse_val_ident(&mdef, endptr);
    skip_white(&mdef, endptr);
  }

  if (mrec->font_name) {
    char *tmp;

    tmp = strip_options(mrec->font_name, &mrec->opt);
    if (tmp) {
      RELEASE(mrec->font_name);
      mrec->font_name = tmp;
    }
  }

  /* Parse any remaining arguments */ 
  while (mdef + 1 < endptr && *mdef == '-') {
    char mopt, *number;

    mopt = mdef[1];
    switch (mopt) {
    case 'b': /* Fake-bold option */
      mdef += 2;
      skip_white(&mdef, endptr);
      number = parse_number(&mdef, endptr);
      if (!number) {
	WARN("Missing bold parameter value...", mrec->tex_name);
	clear_map_record(mrec);
	return -1;
      }
      mrec->opt.bold = atof(number);
      if (mrec->opt.bold < 0.0)
	mrec->opt.bold = 0.0;
      RELEASE(number);
      break;
    case 's': /* Slant option */
      mdef += 2;
      skip_white(&mdef, endptr);
      number = parse_number(&mdef, endptr);
      if (!number) {
	WARN("Missing slant value...", mrec->tex_name);
	clear_map_record(mrec);
	return -1;
      }
      mrec->opt.slant = atof(number);
      RELEASE(number);
      break;
    case 'e': /* Extend option */
      mdef += 2;
      skip_white(&mdef, endptr);
      number = parse_number(&mdef, endptr);
      if (!number) {
	WARN("Missing extend value...", mrec->tex_name);
	clear_map_record(mrec);
	return -1;
      }
      mrec->opt.extend = atof(number);
      RELEASE(number);
      break;
    case 'r': /* Remap option */
      mdef += 2;
      skip_white(&mdef, endptr);
      mrec->opt.flags |= FONTMAP_OPT_REMAP;
      break;
    case 'm': /* Map single bytes char 0xab to double byte char 0xcdab  */
      {
	char *nextptr;
	long  v;

	mdef += 2;
	skip_white(&mdef, endptr);
	if (mdef + 4 <= endptr &&
	    mdef[0] == '<' && mdef[3] == '>') {
	  mdef++;
	  v = strtol(mdef, &nextptr, 16);
	  nextptr++;
	} else {
	  v = strtol(mdef, &nextptr, 0);
	}

	if (nextptr == mdef ||
	    (nextptr < endptr && !isspace(nextptr[0]))) {
	  WARN("Invalid value for option \"-m\": -->%s<--", mdef);
	  clear_map_record(mrec);
	  return -1;
	}
	if (mrec->opt.mapc < 0)
	  mrec->opt.mapc = 0;
	mrec->opt.mapc |= ((v << 8) & 0x0000ff00L);
	mdef = nextptr;
      }
      break;
    case 'l':
      {
	long len;

	mdef += 2;
	skip_white(&mdef, endptr);
	len = 0;
	while (mdef + len < endptr) {
	  if (isspace(mdef[len])) {
	    break;
	  }
	  len++;
	}
	if (mrec->opt.otl_tags)
	  RELEASE(mrec->opt.otl_tags);
	mrec->opt.otl_tags = NEW(len+1, char);

	strcpy(mrec->opt.otl_tags, mdef);
	mdef += len;
      }
      break;
    case 'w': /* Writing mode (for unicode encoding) */
      if (!mrec->enc_name ||
	  strcmp(mrec->enc_name, "unicode"))
	WARN("Fontmap option \"-w\" meaningless for encoding other than \"unicode\".");
      mdef += 2;
      skip_white(&mdef, endptr);
      number = parse_number(&mdef, endptr);
      if (!number) {
	WARN("Missing wmode value...", mrec->tex_name);
	clear_map_record(mrec);
	return -1;
      }
      if (atoi(number) == 1)
	mrec->opt.flags |= FONTMAP_OPT_VERT;
      RELEASE(number);
      break;

    default:
      WARN("Unrecognized font map option: -->%s<--", mdef);
      clear_map_record(mrec);
      return -1;
      break;
    }
    skip_white(&mdef, endptr);
  }
  fill_in_defaults(mrec);

#ifndef WITHOUT_COMPAT
  /*
   * Compatibility to older version:
   *
   * We use strncmp() here to allow UCS2, UniJIS, UCS-GB, etc.
   */
  if (mrec->sfd_name && mrec->enc_name &&
      (!strncmp(mrec->sfd_name, "Uni"  , strlen("Uni"  )) ||
       !strncmp(mrec->sfd_name, "UCS"  , strlen("UCS"  )) ||
       !strncmp(mrec->sfd_name, "UKS"  , strlen("UKS"  )) ||
       !strncmp(mrec->sfd_name, "UGB"  , strlen("UGB"  )) ||
       !strncmp(mrec->sfd_name, "UB5"  , strlen("UB5"  )) ||
       !strncmp(mrec->sfd_name, "UJIS" , strlen("UJIS" )) ||
       !strncmp(mrec->sfd_name, "UBg5" , strlen("UBg5" )) ||
       !strncmp(mrec->sfd_name, "UBig5", strlen("UBig5"))) &&
      (!strcmp(mrec->enc_name, "Identity-H") ||
       !strcmp(mrec->enc_name, "Identity-V"))
      ) {
    if (!mrec->opt.charcoll || !strstr(mrec->opt.charcoll, "UCS")) {
      mrec->opt.charcoll = RENEW(mrec->opt.charcoll,
				 strlen("UCS")+1, char);
      strcpy(mrec->opt.charcoll, "UCS");
    }
  }
#endif /* WITHOUT_COMPAT */
  if (is_new_entry)
    num_fontmap += 1;

  return 0;
}

int
fontmap_add_mapline (char *mline, char *endptr, int mode)
{
  int   ret_val = -1;
  char *tex_name;

  skip_white(&mline, endptr);
  if (mline >= endptr)
    return -1;

  tex_name = parse_ident(&mline, endptr);
  if (!tex_name)
    ret_val = -1;
  else {
    ret_val = fontmap_add_maprec(tex_name, mline, endptr, mode);
    RELEASE(tex_name);
  }

  return ret_val;
}

void
fontmap_read_mapfile (const char *filename, int mode)
{
  FILE *fp;
  char *start = NULL, *end;
  static long  lpos  = 0;

  ASSERT(filename);

  if (verbose)
    MESG("<FONTMAP:%s", filename);

  fp = DPXFOPEN(filename, DPX_RES_TYPE_FONTMAP);
  if (!fp) {
    WARN("Couldn't open the font map file \"%s\".", filename);
    return;
  }

  while ((start = mfgets(work_buffer, WORK_BUFFER_SIZE, fp)) != NULL) {
    end = work_buffer + strlen(work_buffer);
    lpos++;
    skip_white(&start, end);
    if (start >= end || *start == '%')
      continue;
    if (fontmap_add_mapline(start, end, mode) < 0) {
      WARN("Invalid map record in map file \"%s\" (line: %ld).", filename, lpos);
    }
  }
  DPXFCLOSE(fp);

  if (verbose)
    MESG(">");
}

/*
 * fontmap_get_map_record() returns two IDs, map_id and subfont_id if exist.
 */
int
fontmap_get_map_record (const char *tex_name, int *subfont_id)
{
  int    map_id, sfd_id;
  struct map_record *mrec;
  int    base_len;

  *subfont_id = -1;

  if (!fontmap)
    return -1;

  for (map_id = 0; map_id < num_fontmap; map_id++) {
    mrec = GET_MAP_RECORD(map_id);
    if (mrec->deleted)
      continue;
    else if (!mrec->sfd_name) {
      if (!strcmp(mrec->tex_name, tex_name))
	return map_id;
    } else {
      base_len = strlen(mrec->tex_name);
      if (strlen(tex_name) > base_len &&
	  !strncmp(mrec->tex_name, tex_name, base_len)) {
	sfd_id = load_sfd_record(mrec->sfd_name, (char *) tex_name + base_len);
	if (sfd_id >= 0) {
	  *subfont_id = sfd_id;
	  return map_id;
	}
      }
    }
  }

  return -1;
}

#define CHECK_ID(n) if ((n) < 0 || (n) >= num_fontmap) {\
  ERROR("Invalid font map id: %d", (n));\
}

char *
fontmap_tex_name (int map_id)
{
  struct map_record *mrec;

  CHECK_ID(map_id);
  mrec = GET_MAP_RECORD(map_id);

  return mrec->tex_name;
}

char *
fontmap_enc_name (int map_id)
{
  struct map_record *mrec;

  CHECK_ID(map_id);
  mrec = GET_MAP_RECORD(map_id);

  return mrec->enc_name;
}

char *
fontmap_sfd_name (int map_id)
{
  struct map_record *mrec;

  CHECK_ID(map_id);
  mrec = GET_MAP_RECORD(map_id);

  return mrec->sfd_name;
}

char *
fontmap_font_name (int map_id)
{
  struct map_record *mrec;

  CHECK_ID(map_id);
  mrec = GET_MAP_RECORD(map_id);

  return mrec->font_name;
}

double
fontmap_extend (int map_id)
{
  struct map_record *mrec;

  CHECK_ID(map_id);
  mrec = GET_MAP_RECORD(map_id);

  return mrec->opt.extend;
}

double
fontmap_slant (int map_id)
{
  struct map_record *mrec;

  CHECK_ID(map_id);
  mrec = GET_MAP_RECORD(map_id);

  return mrec->opt.slant;
}

double
fontmap_bold (int map_id)
{
  struct map_record *mrec;

  CHECK_ID(map_id);
  mrec = GET_MAP_RECORD(map_id);

  return mrec->opt.bold;
}

int
fontmap_remap (int map_id)
{
  struct map_record *mrec;

  CHECK_ID(map_id);
  mrec = GET_MAP_RECORD(map_id);

  return (int) (mrec->opt.flags & FONTMAP_OPT_REMAP);
}

long
fontmap_mapc (int map_id)
{
  struct map_record *mrec;

  CHECK_ID(map_id);
  mrec = GET_MAP_RECORD(map_id);

  return mrec->opt.mapc;
}

fontmap_opt *
fontmap_get_opt (int map_id)
{
  struct map_record *mrec;

  CHECK_ID(map_id);
  mrec = GET_MAP_RECORD(map_id);

  return &(mrec->opt);
}

/* CIDFont options
 *
 * FORMAT:
 *
 *   (:int:)?!?string(/string)?(,string)?
 */

static char *
substr (char **str, char stop)
{
  char *sstr, *endptr;

  endptr = strchr(*str, stop);
  if (!endptr || endptr == *str)
    return NULL;
  sstr = NEW(endptr-(*str)+1, char);
  memcpy(sstr, *str, endptr-(*str));
  sstr[endptr-(*str)] = '\0';

  *str = endptr+1;
  return sstr;
}

#include <ctype.h>
#define CID_MAPREC_CSI_DELIM '/'

static char *
strip_options (const char *map_name, fontmap_opt *opt)
{
  char *font_name;
  char *p, *next = NULL;
  int   have_csi = 0, have_style = 0;

  ASSERT(opt);

  p = (char *) map_name;
  font_name      = NULL;
  opt->charcoll  = NULL;
  opt->index     = 0;
  opt->style     = FONTMAP_STYLE_NONE;
  opt->flags     = 0;

  if (*p == ':' && isdigit(*(p+1))) {
    opt->index = (int) strtoul(p+1, &next, 10);
    if (*next == ':')
      p = next + 1;      
    else {
      opt->index = 0;
    }
  }
  if (*p == '!') { /* no-embedding */
    if (*(++p) == '\0')
      ERROR("Invalid map record: %s (--> %s)", map_name, p);
    opt->flags |= FONTMAP_OPT_NOEMBED;
  }

  if ((next = strchr(p, CID_MAPREC_CSI_DELIM)) != NULL) {
    if (next == p)
      ERROR("Invalid map record: %s (--> %s)", map_name, p);
    font_name = substr(&p, CID_MAPREC_CSI_DELIM);
    have_csi  = 1;
  } else if ((next = strchr(p, ',')) != NULL) {
    if (next == p)
      ERROR("Invalid map record: %s (--> %s)", map_name, p);
    font_name = substr(&p, ',');
    have_style = 1;
  } else {
    font_name = NEW(strlen(p)+1, char);
    strcpy(font_name, p);
  }

  if (have_csi) {
    if ((next = strchr(p, ',')) != NULL) {
      opt->charcoll = substr(&p, ',');
      have_style = 1;
    } else if (p[0] == '\0') {
      ERROR("Invalid map record: %s.", map_name);
    } else {
      opt->charcoll = NEW(strlen(p)+1, char);
      strcpy(opt->charcoll, p);
    }
  }

  if (have_style) {
    if (!strncmp(p, "BoldItalic", 10)) {
      if (*(p+10))
	ERROR("Invalid map record: %s (--> %s)", map_name, p);
      opt->style = FONTMAP_STYLE_BOLDITALIC;
    } else if (!strncmp(p, "Bold", 4)) {
      if (*(p+4))
	ERROR("Invalid map record: %s (--> %s)", map_name, p);
      opt->style = FONTMAP_STYLE_BOLD;
    } else if (!strncmp(p, "Italic", 6)) {
      if (*(p+6))
	ERROR("Invalid map record: %s (--> %s)", map_name, p);
      opt->style = FONTMAP_STYLE_ITALIC;
    }
  }

  return font_name;
}
